from unittest import TestCase
from terraform_compliance.common.helper import flatten_list, check_port_cidr_ranges
from tests.mocks import MockedData
from copy import deepcopy


class TestHelperFunctions(TestCase):

    def test_flatten_single_dimensional_list(self):
        a = ['a', 'b', 'c']
        b = ['a', 'b', 'c']

        self.assertEqual(flatten_list(a), b)

    def test_flatten_multi_dimensional_list(self):
        a = ['a', 'b', ['c']]
        b = ['a', 'b', 'c']

        self.assertEqual(flatten_list(a), b)

    def test_flatten_multi_dimensional_nested_list(self):
        a = ['a', 'b', ['c', ['d', 'e'], 'f'], 'g', 'h', 'i', ['j', 'k']]
        b = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k']

        self.assertEqual(flatten_list(a), b)

    def test_check_port_cidr_ranges_port_fail(self):
        pass

    def test_check_port_cidr_ranges_port_passed_because_of_different_protocol(self):
        pass

    def test_check_port_cidr_ranges_port_fail_and_protocol_number_is_used(self):
        pass

    def test_check_port_cidr_ranges_port_not_fail_because_of_cidr(self):
        pass

    def test_check_port_cidr_ranges_port_fail_because_of_given_ip_is_a_member_of_cidr(self):
        pass



import re
from netaddr import IPNetwork


# A helper function that will be used to flattan a multi-dimensional multi-nested list
def flatten_list(input):
    new_list = []
    for i in input:
        if type(i) is list:
            i = flatten_list(i)
            for j in i:
                new_list.extend(j)
        else:
            new_list.extend(i)
    return new_list


def generate_target_resource( target_string ):
    target = target_string.strip()
    target = 'resource.{}'.format(target).split('.')

    if target[-1] in ['id', 'name']:
        target.pop(-1)

    return target


def expand_variable(tf_conf, value):
    regex = r'\${var.(.*)\}'
    matches = re.match(regex, value)

    if matches is None:
        return value

    if matches.group(1) not in tf_conf['variable']:
        return value

    return tf_conf['variable'][matches.group(1)]


def check_if_cidr( value ):
    regex = r'(1[0-9][0-9]|2[0-4][0-9]|25[0-5]|[0-9][0-9]|[0-9])\.(1[0-9][0-9]|2[0-4][0-9]|25[0-5]|[0-9][0-9]|[0-9])\.(1[0-9][0-9]|2[0-4][0-9]|25[0-5]|[0-9][0-9]|[0-9])\.(1[0-9][0-9]|2[0-4][0-9]|25[0-5]|[0-9][0-9]|[0-9])\/(3[0-2]|2[0-9]|1[0-9]|[0-9])'
    matches = re.match(regex, value)

    if matches is not None:
        return True

    return False


def is_ip_in_cidr( ip_cidr, cidr ):
    return_value = False
    cidr = eval(cidr)
    if type(cidr) is list:
        for ip_network in cidr:
            if IPNetwork(ip_cidr) in IPNetwork(ip_network):
                return_value = True
    else:
        if IPNetwork(ip_cidr) in IPNetwork(cidr):
            return_value = True

    return return_value


# A helper function that compares port related data with given dictionary
def check_port_cidr_ranges(tf_conf, security_group, proto, port, cidr):
    protocol = ''
    from_port = 0
    to_port = 0
    cidr_blocks = None
    giveError = False

    # This is because of resource_mounting
    #if 'referenced_name' in security_group:
    #    return

    if 'cidr_blocks' in security_group:
        if type(security_group['cidr_blocks']) is list:
            for i in range(0,len(security_group['cidr_blocks'])):
                if not check_if_cidr(security_group['cidr_blocks'][i]):
                    security_group['cidr_blocks'][i] = expand_variable(tf_conf, security_group['cidr_blocks'][i])['default']
        else:
            if not check_if_cidr(security_group['cidr_blocks']):
                security_group['cidr_blocks'] = expand_variable(tf_conf, security_group['cidr_blocks'])['default']

    #TODO: Add custom protocol support where HCL has numbers instead of protocol names like tcp, udp
    #TODO: Add IP Range/Netmask check with the given CIDR.
    for y in security_group:
        if y == 'protocol':
            protocol = [security_group[y]]
            if protocol[0] == '-1':
                protocol = ['tcp', 'udp']

        if y == 'from_port' and security_group[y] > 0:
            from_port = int(security_group[y])

        if y == 'to_port' and security_group[y] > 0:
            to_port = int(security_group[y])

        if y == 'cidr_blocks':
            if type(security_group[y] is list):
                cidr_blocks = str(security_group[y])

    if int(to_port) == 0 and int(from_port) == 0:
        to_port = 65535

    if int(to_port) > int(from_port):
        if int(from_port) <= port <= int(to_port) and proto in protocol and is_ip_in_cidr(cidr, cidr_blocks):
            giveError = True
    elif int(from_port) > int(to_port):
        if int(to_port) <= port <= int(from_port) or proto in protocol and is_ip_in_cidr(cidr, cidr_blocks):
            giveError = True
    elif int(from_port) == int(to_port):
        if int(from_port) == port and proto in protocol and is_ip_in_cidr(cidr, cidr_blocks):
            giveError = True

    if giveError:
        raise AssertionError('Found {}/{} in {}/{}-{} for {}'.format(proto, port, protocol, from_port,
                                                                     to_port, cidr_blocks))


def change_value_in_dict(target_dictionary, path_to_change, value_to_change):
    if type(path_to_change) is str:
        path_to_change = path_to_change.split('.')

    if type(path_to_change) is not list:
        return False

    path_to_adjust = '["{}"]'.format('"]["'.join(path_to_change))
    path_to_check  = '["{}"]["type"]'.format('"]["'.join(path_to_change))
    path_to_add    = '["{}"]'.format('"]["'.join(path_to_change[:-1]))


    try:
        target = eval('target_dictionary{}'.format(path_to_adjust))

        for key, value in value_to_change.items():
            if 'type' in value:
                type_key = value['type']
                source = value
                source['referenced_name'] = key

                if type_key not in target:
                    target[type_key] = list()
                elif type_key in target:
                    if type(target[type_key]) is not list:
                        target[type_key] = [target[type_key]]

                target[type_key].append(source)

        target.update(value_to_change)

        try:
            exec('target_dictionary{}.update({})'.format(path_to_adjust, target))
        except:
            # Yes I know, this is against PEP8.
            pass

    except KeyError:
        pass



# -*- coding: utf-8 -*-

from radish import step, world, custom_type, then, when, given
from terraform_compliance.steps import untaggable_resources, regex, resource_name, encryption_property
from terraform_compliance.common.helper import check_port_cidr_ranges
from terraform_compliance.extensions.terraform_validate import normalise_tag_values


# New Arguments
@custom_type("ANY", r"[\.\/_\-A-Za-z0-9\s]+")
def arg_exp_for_secure_text(text):
    return text

@given(u'I have {resource:ANY} defined')
def define_a_resource(step, resource):
    world.config.terraform.error_if_property_missing()
    if (resource in resource_name.keys()):
        resource = resource_name[resource]

    step.context.resource_type = resource
    step.context.stash = step.context.resources = world.config.terraform.resources(resource)


@step(u'I {action_type:ANY} them')
def i_action_them(step, action_type):
    if not step.context.resources.resource_list:
        return

    if action_type == "count":
        step.context.stash = len(step.context.stash.resource_list)
    elif action_type == "sum":
        step.context.stash = sum(step.context.stash.resource_list)
    else:
        AssertionError("Invalid action_type in the scenario: {}".format(action))


@step(u'I expect the result is {operator:ANY} than {number:d}')
def func(step, operator, number):
    if not step.context.resources.resource_list:
        return

    value = int(step.context.stash)

    if operator == "more":
        assert value > number, str(value) + " is not more than " + str(number)
    elif operator == "more and equal":
        assert value >= number, str(value) + " is not more and equal than " + str(number)
    elif operator == "less":
        assert value < number, str(value) + " is not less than " + str(number)
    elif operator == "less and equal":
        assert value <= number, str(value) + " is not less and equal than " + str(number)
    else:
        AssertionError("Invalid operator: " + str(operator))


@step(u'it {condition:ANY} contain {something:ANY}')
def func(step, condition, something):
    if not step.context.resources.resource_list:
        return

    if condition == 'must':
        world.config.terraform.error_if_property_missing()

    if something in resource_name.keys():
        something = resource_name[something]

    step.context.resource_type = something
    step.context.resources = step.context.resources.property(something)

    if condition == 'must':
        assert step.context.resources.properties


@step(u'encryption is enabled')
@step(u'encryption must be enabled')
def func(step):
    if not step.context.resources.resource_list:
        return

    world.config.terraform.error_if_property_missing()
    prop = encryption_property[step.context.resource_type]
    step.context.resources.property(prop).should_equal(True)


@step(u'its value must match the "{regex_type}" regex')
def func(step, regex_type):
    if not step.context.resources.resource_list:
        return

    normalise_tag_values(step.context.properties)
    step.context.properties.property(regex_type).should_match_regex(step.context.regex)


@step(u'its value must be set by a variable')
def func(step):
    if not step.context.resources.resource_list:
        return

    step.context.resources.property(step.context.search_value).should_match_regex('\${var.(.*)}')


@step(u'it must not have {proto} protocol and port {port:d} for {cidr:ANY}')
def func(step, proto, port, cidr):
    if not step.context.resources.resource_list:
        return

    proto = str(proto)
    port = int(port)
    cidr = str(cidr)

    for item in step.context.resources.properties:
        if type(item.property_value) is list:
            for security_group in item.property_value:
                check_port_cidr_ranges(world.config.terraform.terraform_config, security_group, proto, port, cidr)
        else:
            check_port_cidr_ranges(world.config.terraform.terraform_config, item.property_value, proto, port, cidr)